# unit tests for ansible other facter fact collector
# -*- coding: utf-8 -*-
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from unittest.mock import Mock, patch

from .. base import BaseFactsTest

from ansible.module_utils.facts.other.facter import FacterFactCollector

facter_json_output = '''
{
  "operatingsystemmajrelease": "25",
  "hardwareisa": "x86_64",
  "kernel": "Linux",
  "path": "/home/testuser/src/ansible/bin:/home/testuser/perl5/bin:/home/testuser/perl5/bin:/home/testuser/bin:/home/testuser/.local/bin:/home/testuser/pythons/bin:/usr/lib64/qt-3.3/bin:/usr/local/bin:/usr/bin:/usr/local/sbin:/usr/sbin:/home/testuser/.cabal/bin:/home/testuser/gopath/bin:/home/testuser/.rvm/bin",
  "memorysize": "15.36 GB",
  "memoryfree": "4.88 GB",
  "swapsize": "7.70 GB",
  "swapfree": "6.75 GB",
  "swapsize_mb": "7880.00",
  "swapfree_mb": "6911.41",
  "memorysize_mb": "15732.95",
  "memoryfree_mb": "4997.68",
  "lsbmajdistrelease": "25",
  "macaddress": "02:42:ea:15:d8:84",
  "id": "testuser",
  "domain": "example.com",
  "augeasversion": "1.7.0",
  "os": {
    "name": "Fedora",
    "family": "RedHat",
    "release": {
      "major": "25",
      "full": "25"
    },
    "lsb": {
      "distcodename": "TwentyFive",
      "distid": "Fedora",
      "distdescription": "Fedora release 25 (Twenty Five)",
      "release": ":core-4.1-amd64:core-4.1-noarch:cxx-4.1-amd64:cxx-4.1-noarch:desktop-4.1-amd64:desktop-4.1-noarch:languages-4.1-amd64:languages-4.1-noarch:printing-4.1-amd64:printing-4.1-noarch",
      "distrelease": "25",
      "majdistrelease": "25"
    }
  },
  "processors": {
    "models": [
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
      "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz"
    ],
    "count": 8,
    "physicalcount": 1
  },
  "architecture": "x86_64",
  "hardwaremodel": "x86_64",
  "operatingsystem": "Fedora",
  "processor0": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor1": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor2": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor3": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor4": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor5": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor6": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processor7": "Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz",
  "processorcount": 8,
  "uptime_seconds": 1558090,
  "fqdn": "myhostname.example.com",
  "rubyversion": "2.3.3",
  "gid": "testuser",
  "physicalprocessorcount": 1,
  "netmask": "255.255.0.0",
  "uniqueid": "a8c01301",
  "uptime_days": 18,
  "interfaces": "docker0,em1,lo,vethf20ff12,virbr0,virbr1,virbr0_nic,virbr1_nic,wlp4s0",
  "ipaddress_docker0": "172.17.0.1",
  "macaddress_docker0": "02:42:ea:15:d8:84",
  "netmask_docker0": "255.255.0.0",
  "mtu_docker0": 1500,
  "macaddress_em1": "3c:97:0e:e9:28:8e",
  "mtu_em1": 1500,
  "ipaddress_lo": "127.0.0.1",
  "netmask_lo": "255.0.0.0",
  "mtu_lo": 65536,
  "macaddress_vethf20ff12": "ae:6e:2b:1e:a1:31",
  "mtu_vethf20ff12": 1500,
  "ipaddress_virbr0": "192.168.137.1",
  "macaddress_virbr0": "52:54:00:ce:82:5e",
  "netmask_virbr0": "255.255.255.0",
  "mtu_virbr0": 1500,
  "ipaddress_virbr1": "192.168.121.1",
  "macaddress_virbr1": "52:54:00:b4:68:a9",
  "netmask_virbr1": "255.255.255.0",
  "mtu_virbr1": 1500,
  "macaddress_virbr0_nic": "52:54:00:ce:82:5e",
  "mtu_virbr0_nic": 1500,
  "macaddress_virbr1_nic": "52:54:00:b4:68:a9",
  "mtu_virbr1_nic": 1500,
  "ipaddress_wlp4s0": "192.168.1.19",
  "macaddress_wlp4s0": "5c:51:4f:e6:a8:e3",
  "netmask_wlp4s0": "255.255.255.0",
  "mtu_wlp4s0": 1500,
  "virtual": "physical",
  "is_virtual": false,
  "partitions": {
    "sda2": {
      "size": "499091456"
    },
    "sda1": {
      "uuid": "32caaec3-ef40-4691-a3b6-438c3f9bc1c0",
      "size": "1024000",
      "mount": "/boot"
    }
  },
  "lsbdistcodename": "TwentyFive",
  "lsbrelease": ":core-4.1-amd64:core-4.1-noarch:cxx-4.1-amd64:cxx-4.1-noarch:desktop-4.1-amd64:desktop-4.1-noarch:languages-4.1-amd64:languages-4.1-noarch:printing-4.1-amd64:printing-4.1-noarch",  # noqa
  "filesystems": "btrfs,ext2,ext3,ext4,xfs",
  "system_uptime": {
    "seconds": 1558090,
    "hours": 432,
    "days": 18,
    "uptime": "18 days"
  },
  "ipaddress": "172.17.0.1",
  "timezone": "EDT",
  "ps": "ps -ef",
  "rubyplatform": "x86_64-linux",
  "rubysitedir": "/usr/local/share/ruby/site_ruby",
  "uptime": "18 days",
  "lsbdistrelease": "25",
  "operatingsystemrelease": "25",
  "facterversion": "2.4.3",
  "kernelrelease": "4.9.14-200.fc25.x86_64",
  "lsbdistdescription": "Fedora release 25 (Twenty Five)",
  "network_docker0": "172.17.0.0",
  "network_lo": "127.0.0.0",
  "network_virbr0": "192.168.137.0",
  "network_virbr1": "192.168.121.0",
  "network_wlp4s0": "192.168.1.0",
  "lsbdistid": "Fedora",
  "selinux": true,
  "selinux_enforced": false,
  "selinux_policyversion": "30",
  "selinux_current_mode": "permissive",
  "selinux_config_mode": "permissive",
  "selinux_config_policy": "targeted",
  "hostname": "myhostname",
  "osfamily": "RedHat",
  "kernelmajversion": "4.9",
  "blockdevice_sr0_size": 1073741312,
  "blockdevice_sr0_vendor": "MATSHITA",
  "blockdevice_sr0_model": "DVD-RAM UJ8E2",
  "blockdevice_sda_size": 256060514304,
  "blockdevice_sda_vendor": "ATA",
  "blockdevice_sda_model": "SAMSUNG MZ7TD256",
  "blockdevices": "sda,sr0",
  "uptime_hours": 432,
  "kernelversion": "4.9.14"
}
'''


class TestFacterCollector(BaseFactsTest):
    __test__ = True
    gather_subset = ['!all', 'facter']
    valid_subsets = ['facter']
    fact_namespace = 'ansible_facter'
    collector_class = FacterFactCollector

    def _mock_module(self):
        mock_module = Mock()
        mock_module.params = {'gather_subset': self.gather_subset,
                              'gather_timeout': 10,
                              'filter': '*'}
        mock_module.get_bin_path = Mock(return_value='/not/actually/facter')
        mock_module.run_command = Mock(return_value=(0, facter_json_output, ''))
        return mock_module

    @patch('ansible.module_utils.facts.other.facter.FacterFactCollector.get_facter_output')
    def test_bogus_json(self, mock_get_facter_output):
        module = self._mock_module()

        # bogus json
        mock_get_facter_output.return_value = '{'

        fact_collector = self.collector_class()
        facts_dict = fact_collector.collect(module=module)

        self.assertIsInstance(facts_dict, dict)
        self.assertEqual(facts_dict, {})

    @patch('ansible.module_utils.facts.other.facter.FacterFactCollector.run_facter')
    def test_facter_non_zero_return_code(self, mock_run_facter):
        module = self._mock_module()

        # bogus json
        mock_run_facter.return_value = (1, '{}', '')

        fact_collector = self.collector_class()
        facts_dict = fact_collector.collect(module=module)

        self.assertIsInstance(facts_dict, dict)

        # This assumes no 'facter' entry at all is correct
        self.assertNotIn('facter', facts_dict)
        self.assertEqual(facts_dict, {})



from __future__ import annotations

from ansible.plugins.vars import BaseVarsPlugin


class VarsModule(BaseVarsPlugin):
    REQUIRES_WHITELIST = True

    def get_vars(self, loader, path, entities):
        return {}



# (c) 2015, Marius Gedminas <marius@gedmin.as>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from ansible.utils.helpers import pct_to_int


class TestHelpers(unittest.TestCase):

    def test_pct_to_int(self):
        self.assertEqual(pct_to_int(1, 100), 1)
        self.assertEqual(pct_to_int(-1, 100), -1)
        self.assertEqual(pct_to_int("1%", 10), 1)
        self.assertEqual(pct_to_int("1%", 10, 0), 0)
        self.assertEqual(pct_to_int("1", 100), 1)
        self.assertEqual(pct_to_int("10%", 100), 10)



#!/usr/bin/python

from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
# overridden
from ansible.module_utils.ansible_release import data

results = {"data": data}

arg_spec = dict(
    foo=dict(type='str', aliases=['baz'], deprecated_aliases=[dict(name='baz', version='9.99')])
)

AnsibleModule(argument_spec=arg_spec).exit_json(**results)



"""Support code for working without a supported CI provider."""
from __future__ import annotations

import os
import platform
import random
import re
import typing as t

from ..config import (
    CommonConfig,
    TestConfig,
)

from ..io import (
    read_text_file,
)

from ..git import (
    Git,
)

from ..util import (
    ApplicationError,
    display,
    is_binary_file,
    SubprocessError,
)

from . import (
    CIProvider,
)

CODE = ''  # not really a CI provider, so use an empty string for the code


class Local(CIProvider):
    """CI provider implementation when not using CI."""

    priority = 1000

    @staticmethod
    def is_supported() -> bool:
        """Return True if this provider is supported in the current running environment."""
        return True

    @property
    def code(self) -> str:
        """Return a unique code representing this provider."""
        return CODE

    @property
    def name(self) -> str:
        """Return descriptive name for this provider."""
        return 'Local'

    def generate_resource_prefix(self) -> str:
        """Return a resource prefix specific to this CI provider."""
        prefix = 'ansible-test-%d-%s' % (
            random.randint(10000000, 99999999),
            platform.node().split('.')[0],
        )

        return prefix

    def get_base_commit(self, args: CommonConfig) -> str:
        """Return the base commit or an empty string."""
        return ''

    def detect_changes(self, args: TestConfig) -> t.Optional[list[str]]:
        """Initialize change detection."""
        result = LocalChanges(args)

        display.info('Detected branch %s forked from %s at commit %s' % (
            result.current_branch, result.fork_branch, result.fork_point))

        if result.untracked and not args.untracked:
            display.warning('Ignored %s untracked file(s). Use --untracked to include them.' %
                            len(result.untracked))

        if result.committed and not args.committed:
            display.warning('Ignored %s committed change(s). Omit --ignore-committed to include them.' %
                            len(result.committed))

        if result.staged and not args.staged:
            display.warning('Ignored %s staged change(s). Omit --ignore-staged to include them.' %
                            len(result.staged))

        if result.unstaged and not args.unstaged:
            display.warning('Ignored %s unstaged change(s). Omit --ignore-unstaged to include them.' %
                            len(result.unstaged))

        names = set()

        if args.tracked:
            names |= set(result.tracked)
        if args.untracked:
            names |= set(result.untracked)
        if args.committed:
            names |= set(result.committed)
        if args.staged:
            names |= set(result.staged)
        if args.unstaged:
            names |= set(result.unstaged)

        if not args.metadata.changes:
            args.metadata.populate_changes(result.diff)

            for path in result.untracked:
                if is_binary_file(path):
                    args.metadata.changes[path] = ((0, 0),)
                    continue

                line_count = len(read_text_file(path).splitlines())

                args.metadata.changes[path] = ((1, line_count),)

        return sorted(names)

    def supports_core_ci_auth(self) -> bool:
        """Return True if Ansible Core CI is supported."""
        path = self._get_aci_key_path()
        return os.path.exists(path)

    def prepare_core_ci_auth(self) -> dict[str, t.Any]:
        """Return authentication details for Ansible Core CI."""
        path = self._get_aci_key_path()
        auth_key = read_text_file(path).strip()

        request = dict(
            key=auth_key,
            nonce=None,
        )

        auth = dict(
            remote=request,
        )

        return auth

    def get_git_details(self, args: CommonConfig) -> t.Optional[dict[str, t.Any]]:
        """Return details about git in the current environment."""
        return None  # not yet implemented for local

    @staticmethod
    def _get_aci_key_path() -> str:
        path = os.path.expanduser('~/.ansible-core-ci.key')
        return path


class InvalidBranch(ApplicationError):
    """Exception for invalid branch specification."""

    def __init__(self, branch: str, reason: str) -> None:
        message = 'Invalid branch: %s\n%s' % (branch, reason)

        super().__init__(message)

        self.branch = branch


class LocalChanges:
    """Change information for local work."""

    def __init__(self, args: TestConfig) -> None:
        self.args = args
        self.git = Git()

        self.current_branch = self.git.get_branch()

        if self.is_official_branch(self.current_branch):
            raise InvalidBranch(branch=self.current_branch,
                                reason='Current branch is not a feature branch.')

        self.fork_branch = None
        self.fork_point = None

        self.local_branches = sorted(self.git.get_branches())
        self.official_branches = sorted([b for b in self.local_branches if self.is_official_branch(b)])

        for self.fork_branch in self.official_branches:
            try:
                self.fork_point = self.git.get_branch_fork_point(self.fork_branch)
                break
            except SubprocessError:
                pass

        if self.fork_point is None:
            raise ApplicationError('Unable to auto-detect fork branch and fork point.')

        # tracked files (including unchanged)
        self.tracked = sorted(self.git.get_file_names(['--cached']))
        # untracked files (except ignored)
        self.untracked = sorted(self.git.get_file_names(['--others', '--exclude-standard']))
        # tracked changes (including deletions) committed since the branch was forked
        self.committed = sorted(self.git.get_diff_names([self.fork_point, 'HEAD']))
        # tracked changes (including deletions) which are staged
        self.staged = sorted(self.git.get_diff_names(['--cached']))
        # tracked changes (including deletions) which are not staged
        self.unstaged = sorted(self.git.get_diff_names([]))
        # diff of all tracked files from fork point to working copy
        self.diff = self.git.get_diff([self.fork_point])

    def is_official_branch(self, name: str) -> bool:
        """Return True if the given branch name an official branch for development or releases."""
        if self.args.base_branch:
            return name == self.args.base_branch

        if name == 'devel':
            return True

        if re.match(r'^stable-[0-9]+\.[0-9]+$', name):
            return True

        return False



# (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = '''
    name: display_resolved_action
    type: aggregate
    short_description: Displays the requested and resolved actions at the end of a playbook.
    description:
        - Displays the requested and resolved actions in the format "requested == resolved".
    requirements:
      - Enable in configuration.
'''

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'display_resolved_action'
    CALLBACK_NEEDS_ENABLED = True

    def __init__(self, *args, **kwargs):
        super(CallbackModule, self).__init__(*args, **kwargs)
        self.requested_to_resolved = {}

    def v2_playbook_on_task_start(self, task, is_conditional):
        self.requested_to_resolved[task.action] = task.resolved_action

    def v2_playbook_on_stats(self, stats):
        for requested, resolved in self.requested_to_resolved.items():
            self._display.display("%s == %s" % (requested, resolved), screen_only=True)



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright (c), Michael DeHaan <michael.dehaan@gmail.com>, 2012-2013
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations


def _get_quote_state(token, quote_char):
    '''
    the goal of this block is to determine if the quoted string
    is unterminated in which case it needs to be put back together
    '''
    # the char before the current one, used to see if
    # the current character is escaped
    prev_char = None
    for idx, cur_char in enumerate(token):
        if idx > 0:
            prev_char = token[idx - 1]
        if cur_char in '"\'' and prev_char != '\\':
            if quote_char:
                if cur_char == quote_char:
                    quote_char = None
            else:
                quote_char = cur_char
    return quote_char


def _count_jinja2_blocks(token, cur_depth, open_token, close_token):
    '''
    this function counts the number of opening/closing blocks for a
    given opening/closing type and adjusts the current depth for that
    block based on the difference
    '''
    num_open = token.count(open_token)
    num_close = token.count(close_token)
    if num_open != num_close:
        cur_depth += (num_open - num_close)
        if cur_depth < 0:
            cur_depth = 0
    return cur_depth


def split_args(args):
    '''
    Splits args on whitespace, but intelligently reassembles
    those that may have been split over a jinja2 block or quotes.

    When used in a remote module, we won't ever have to be concerned about
    jinja2 blocks, however this function is/will be used in the
    core portions as well before the args are templated.

    example input: a=b c="foo bar"
    example output: ['a=b', 'c="foo bar"']

    Basically this is a variation shlex that has some more intelligence for
    how Ansible needs to use it.
    '''

    # the list of params parsed out of the arg string
    # this is going to be the result value when we are donei
    params = []

    # here we encode the args, so we have a uniform charset to
    # work with, and split on white space
    args = args.strip()
    try:
        args = args.encode('utf-8')
        do_decode = True
    except UnicodeDecodeError:
        do_decode = False
    items = args.split('\n')

    # iterate over the tokens, and reassemble any that may have been
    # split on a space inside a jinja2 block.
    # ex if tokens are "{{", "foo", "}}" these go together

    # These variables are used
    # to keep track of the state of the parsing, since blocks and quotes
    # may be nested within each other.

    quote_char = None
    inside_quotes = False
    print_depth = 0  # used to count nested jinja2 {{ }} blocks
    block_depth = 0  # used to count nested jinja2 {% %} blocks
    comment_depth = 0  # used to count nested jinja2 {# #} blocks

    # now we loop over each split chunk, coalescing tokens if the white space
    # split occurred within quotes or a jinja2 block of some kind
    for itemidx, item in enumerate(items):

        # we split on spaces and newlines separately, so that we
        # can tell which character we split on for reassembly
        # inside quotation characters
        tokens = item.strip().split(' ')

        line_continuation = False
        for idx, token in enumerate(tokens):

            # if we hit a line continuation character, but
            # we're not inside quotes, ignore it and continue
            # on to the next token while setting a flag
            if token == '\\' and not inside_quotes:
                line_continuation = True
                continue

            # store the previous quoting state for checking later
            was_inside_quotes = inside_quotes
            quote_char = _get_quote_state(token, quote_char)
            inside_quotes = quote_char is not None

            # multiple conditions may append a token to the list of params,
            # so we keep track with this flag to make sure it only happens once
            # append means add to the end of the list, don't append means concatenate
            # it to the end of the last token
            appended = False

            # if we're inside quotes now, but weren't before, append the token
            # to the end of the list, since we'll tack on more to it later
            # otherwise, if we're inside any jinja2 block, inside quotes, or we were
            # inside quotes (but aren't now) concat this token to the last param
            if inside_quotes and not was_inside_quotes:
                params.append(token)
                appended = True
            elif print_depth or block_depth or comment_depth or inside_quotes or was_inside_quotes:
                if idx == 0 and not inside_quotes and was_inside_quotes:
                    params[-1] = "%s%s" % (params[-1], token)
                elif len(tokens) > 1:
                    spacer = ''
                    if idx > 0:
                        spacer = ' '
                    params[-1] = "%s%s%s" % (params[-1], spacer, token)
                else:
                    spacer = ''
                    if not params[-1].endswith('\n') and idx == 0:
                        spacer = '\n'
                    params[-1] = "%s%s%s" % (params[-1], spacer, token)
                appended = True

            # if the number of paired block tags is not the same, the depth has changed, so we calculate that here
            # and may append the current token to the params (if we haven't previously done so)
            prev_print_depth = print_depth
            print_depth = _count_jinja2_blocks(token, print_depth, "{{", "}}")
            if print_depth != prev_print_depth and not appended:
                params.append(token)
                appended = True

            prev_block_depth = block_depth
            block_depth = _count_jinja2_blocks(token, block_depth, "{%", "%}")
            if block_depth != prev_block_depth and not appended:
                params.append(token)
                appended = True

            prev_comment_depth = comment_depth
            comment_depth = _count_jinja2_blocks(token, comment_depth, "{#", "#}")
            if comment_depth != prev_comment_depth and not appended:
                params.append(token)
                appended = True

            # finally, if we're at zero depth for all blocks and not inside quotes, and have not
            # yet appended anything to the list of params, we do so now
            if not (print_depth or block_depth or comment_depth) and not inside_quotes and not appended and token != '':
                params.append(token)

        # if this was the last token in the list, and we have more than
        # one item (meaning we split on newlines), add a newline back here
        # to preserve the original structure
        if len(items) > 1 and itemidx != len(items) - 1 and not line_continuation:
            if not params[-1].endswith('\n') or item == '':
                params[-1] += '\n'

        # always clear the line continuation flag
        line_continuation = False

    # If we're done and things are not at zero depth or we're still inside quotes,
    # raise an error to indicate that the args were unbalanced
    if print_depth or block_depth or comment_depth or inside_quotes:
        raise Exception("error while splitting arguments, either an unbalanced jinja2 block or quotes")

    # finally, we decode each param back to the unicode it was in the arg string
    if do_decode:
        params = [x.decode('utf-8') for x in params]

    return params


def is_quoted(data):
    return len(data) > 0 and (data[0] == '"' and data[-1] == '"' or data[0] == "'" and data[-1] == "'")


def unquote(data):
    ''' removes first and last quotes from a string, if the string starts and ends with the same quotes '''
    if is_quoted(data):
        return data[1:-1]
    return data
